/* This file is part of Chunky Monkey IRCD 
 *
 * Chunky Monkey IRCD is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Chunky Monkey IRCD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Chunky Monkey IRCD; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * $Id: ircd.c,v 1.56.4.38 2002/06/05 11:24:24 lloydy Exp $
 *
 */
#include "ircd.h"
#include "chunkyircd.h"
#include "commands.h"
#include "parse.h"
#include "temp.h"
#include "modules.h"
#include "temp_server.h"
#include "temp_client.h"
#include "resolver.h"

/* Describe The Server State */
int signal_caught = 0;		/* Addressed in main.c and modules.c */
static rlim_t max_fds;
static struct fd_wrap *fd_list = NULL;
static int highest_fd;

/* Utility routines */
static int ircd_init(struct bind_info [], fd_set *, fd_set *);
static int ircd_sockets_check(struct bind_info [], fd_set *, fd_set *);
/* Accept Connections */
static int ircd_newconnection(struct bind_info *);
/* Deal with commands */
static int ircd_command(int);
static int ircd_temp_command(struct tempinfo *, Cmd);
static int ircd_regist_command(int, Cmd);
/* Register Clients/Servers */
static int ircd_client_register(struct tempinfo *);
static int ircd_server_register(struct tempinfo *);
/* Scan stuff */
static int ircd_check_ping(struct fd_wrap *, int);
static int ircd_check_resolver(struct fd_wrap *);
/* Relays numerics if required */
static int ircd_numeric_relay(Cmd);


/* 
* Main entry point 
*/
int serve_clients(struct bind_info server_sockets[])
{
	unsigned int i;
	struct timeval timeout;
	fd_set connections;
	fd_set serv_connects;
	time_t last_pingscan;
	time_t last_check;
	int should_ping = 0;

	/* Set everything up */
	ircd_init(server_sockets, &connections, &serv_connects);

	/* The timeout needs setting */
	timeout.tv_sec = 1;
	timeout.tv_usec = 0;

	last_pingscan = time(NULL);
	last_check = time(NULL);

	/* The while loop which is used for serving */
	while (signal_caught == 0) {

		/* call select() */
		select(highest_fd + 1, &connections, &serv_connects, NULL,
		       &timeout);

		/* Check sockets for stuff */
		ircd_sockets_check(server_sockets, &connections,
				   &serv_connects);


		/* Check pings every 30 seconds */
		if ((last_pingscan + 30) < time(NULL)) {
			last_pingscan = time(NULL);
			should_ping = 1;
		} else {
			should_ping = 0;
		}

		/* Reset the fd_set */
		FD_ZERO(&connections);
		FD_ZERO(&serv_connects);
		for (i = 0; i < 32 && server_sockets[i].fd != -1; i++)
			FD_SET(server_sockets[i].fd, &connections);

		/* Scan the array for regist/ping checks */
		for (i = 0; i < max_fds; i++) {

			/* Ping if we should */
			if (should_ping) {
				ircd_check_ping(fd_list, i);
			}

			/* Check the resolver */
			ircd_check_resolver(&fd_list[i]);

			/* Try and Register the client */
			if (fd_list[i].temp) {
				if (!ircd_client_register(fd_list[i].temp)) {
					ircd_server_register(fd_list[i].
							     temp);
				}
			}

			/* Now they're in right state (Client/Server/Temp */
			/* Add to fd list */
			if (fd_list[i].temp) {

				/* Add to pending server connects() */
				if (fd_list[i].temp->flags & USR_CONNECT) {
					FD_SET(i, &serv_connects);
				}
				/* Only add if it isn't waiting for a resolv return */
				if (!
				    (fd_list[i].temp->
				     flags & USR_SHOULDCONN)) {
					FD_SET(i, &connections);
				}
			} else {
				if (fd_list[i].server) {
					FD_SET(i, &connections);
				}

				if (fd_list[i].client) {
					FD_SET(i, &connections);
				}
			}
		}

		/* Reset highest_fd */
		if (highest_fd == -1) {
			for (i = 0; i < 32 && server_sockets[i].fd != -1; i++)
				highest_fd = server_sockets[i].fd;
		}

		/* Reset timeval */
		timeout.tv_sec = 1;
		timeout.tv_usec = 0;
	}

	return 1;
}

/* Check the state of the various sockets in the system */
/* Call appropriate stuff */
static int ircd_sockets_check(struct bind_info server_sockets[], fd_set * connections,
			      fd_set * serv_connects)
{
	unsigned int i;
	int retval;
	unsigned int rv_size = sizeof(int);

	/* Check for a new connection */
	for (i = 0; i < 32 && server_sockets[i].fd != -1; i++) {
		if (FD_ISSET(server_sockets[i].fd, connections)) {
			/* Accept new connection */
			return ircd_newconnection(&server_sockets[i]);
		}
	}

	/* Scan array for client/server/temp pointer */
	for (i = 0; i < max_fds; i++) {
		if (FD_ISSET(i, connections)) {
			/* Deal with it */
			return ircd_command(i);
		} else {
			/* We've returned due to an event on a pending CONNECT */
			if (FD_ISSET(i, serv_connects)) {
				log_msg(LOG_DEBUG2,
					"Running getsockopt() on socket %d",
					i);
				if (!getsockopt
				    (i, SOL_SOCKET, SO_ERROR,
				     (void *) &retval, &rv_size)) {
					if (retval != 0) {
						log_msg(LOG_DEBUG2,
							"Connect call failed, with error %s",
							strerror(retval));
						/* Remove flags */
						temp_rem(i);
						fd_list[i].temp = NULL;
						close(i);
						return 0;
					} else {
						/* The socket is connected, init the connection */
						log_msg(LOG_DEBUG2,
							"Connected on fd %d, calling init_server_connection");
						return server_init_link(fd_list[i].temp);
					}
				} else {
					return 0;
				}
			}
		}
	}

	return 1;
}


/* 
 * Accept a new connection, add to temp list 
 */
static int ircd_newconnection(struct bind_info *server_sock)
{
	struct sockaddr_in client_info;
	int client_fd = -1;
	int sin_size;

	/* Find size of sockaddr */
	sin_size = sizeof(struct sockaddr);

	/* Accept the connection */
	client_fd =
	    accept(server_sock->fd, (struct sockaddr *) &client_info,
		   (size_t *) & sin_size);
	if (client_fd == -1) {
		log_msg(LOG_DEBUG1,
			"Failed to accept connection, reason %s",
			strerror(errno));
		return 0;
	}

	/* Set non-blocking */
	fcntl(client_fd, F_SETFL, O_NONBLOCK);

	/* Update highest fd */
	if (client_fd > highest_fd)
		highest_fd = client_fd;

	/* Call temp_add */
	/* Add to fd array */
	fd_list[client_fd].temp =
	    temp_add((char *) inet_ntoa(client_info.sin_addr), client_fd);

	if (fd_list[client_fd].temp == NULL) {
		/* Shite!!! */
		log_msg(LOG_DEBUG1,
			"Failed to add client into temp array");
		/* Drop client */
		close(client_fd);
		return 0;
	}

	fd_list[client_fd].temp->lastrecv = time(NULL);
	fd_list[client_fd].res_id =
	    resolver_submit_reverse(client_info.sin_addr.s_addr);
	
#ifdef USE_OPENSSL
	if(server_sock->use_ssl) {
		fd_list[client_fd].temp->ssl = SSL_new(server_config.ssl_ctx);
		assert(fd_list[client_fd].temp->ssl != NULL);

		SSL_set_fd(fd_list[client_fd].temp->ssl,
				fd_list[client_fd].temp->fd);

		SSL_accept(fd_list[client_fd].temp->ssl);
	}
#endif


	return 1;

}

/* 
 * Recieve data, parse it, and decide what to do with it 
 */
static int ircd_command(int client_fd)
{
	char buffer[2048];
	Cmd cmd = NULL;
	int numbytes;

	errno = 0;

	memset(buffer, 0, 2048);

	while ((numbytes = ircd_recv(ircd_fdlist_get_entry(client_fd),buffer)) > 0) {

		/* Flood protection */
		if (fd_list[client_fd].client != NULL && numbytes > 0) {

			if (fd_list[client_fd].client->flood_timer <
			    time(NULL)) {
				fd_list[client_fd].client->flood_timer =
				    time(NULL);
			} else {
				fd_list[client_fd].client->flood_timer +=
				    2;
			}

			if (fd_list[client_fd].client->flood_timer >
			    time(NULL) + 30) {
				/* They're heavily flooding us, drop them */
				force_quit(client_fd, "Excess Flood");
				return 0;
			}

			if (fd_list[client_fd].client->flood_timer >
			    time(NULL) + 15) {
				/* Slight flood, ignore them */
				return 0;
			}
		}

		/* Touch the idle timers */
		if (fd_list[client_fd].client) {
			fd_list[client_fd].client->lastrecv = time(NULL);
		}
		if (fd_list[client_fd].server) {
			fd_list[client_fd].server->lastrecv = time(NULL);
		}
		if (fd_list[client_fd].temp) {
			fd_list[client_fd].temp->lastrecv = time(NULL);
		}

		/* Call the Parser */
		/* We have to loop here, blame SSL */
		cmd = parse_command(buffer);

		if (fd_list[client_fd].temp) {
			/* Clients in the temp list are restricted */
			ircd_temp_command(fd_list[client_fd].temp, cmd);
		} else {
			ircd_regist_command(client_fd, cmd);
		} 
		
	}

	/* If we had a socket error, kill client */
	if (numbytes <= 0 || (errno)) {

		if(errno == EAGAIN)
			return 0;

		/* This is wierd, handle connection reset */
		if(numbytes == 0) 
			errno = ECONNRESET;

		force_quit(client_fd, strerror(errno));
	}

	/* Else feck off */
	return 0;



}

/* 
* Deal with a command from a registered client/server 
*/
static int ircd_regist_command(int fd, Cmd cmd)
{
	char buffer[BUF_SIZE];

	/* Unload is a special command */
	if (lc_strncmp(cmd->command, "unload", -1) == 0) {
		cmd_unload(fd_list[fd].client, fd_list[fd].server, cmd);
		cmd_free(cmd);
		return 1;
	}

	/* So is REHASH */
	if (lc_strncmp(cmd->command, "rehash", -1) == 0) {
		cmd_rehash(fd_list[fd].client, fd_list[fd].server, cmd);
		cmd_free(cmd);
		return 1;
	}

	/* Do other commands */
	if (ircd_module_dlopen
	    (fd_list[fd].client, fd_list[fd].server, cmd)) {
		cmd_free(cmd);
		return 1;
	}

	if (fd_list[fd].server) {
		/* Try and relay a numeric */
		return ircd_numeric_relay(cmd);
	}

	/* Invalid command return */
	if (fd_list[fd].client) {
		/* Shouldn't get one from another server */
		ircd_sprintf(buffer, BUF_SIZE, "%s :Unknown Command",
			     cmd->command);
		send_numeric(fd_list[fd].client, get_my_server(), 421,
			     buffer);
	}

	cmd_free(cmd);
	return 1;

}


/* 
* Deal with a command from a client in the temp list 
*/
static int ircd_temp_command(struct tempinfo *temp_ptr, Cmd cmd)
{
	char buffer[BUF_SIZE];
	char *hostname = NULL;

	if (lc_strncmp(cmd->command, "nick", -1) == 0) {
		temp_nick(temp_ptr, cmd);
		cmd_free(cmd);
		return 1;
	}
	if (lc_strncmp(cmd->command, "user", -1) == 0) {
		temp_user(temp_ptr, cmd);
		cmd_free(cmd);
		return 1;
	}
	if (lc_strncmp(cmd->command, "quit", -1) == 0) {
		temp_quit(temp_ptr, cmd);
		cmd_free(cmd);
		return 1;
	}

	/* Server registration commands */
	/* Check resolver, so that the reverse works better */
	if (fd_list[temp_ptr->fd].res_id != -1) {
		hostname = resolver_check(fd_list[temp_ptr->fd].res_id);
		if (hostname) {
			strncpy(temp_ptr->host, hostname, HOST_LENGTH);
			temp_ptr->flags |= USR_RESOLVED;
		}
	}

	if (lc_strncmp(cmd->command, "server", -1) == 0) {
		temp_server(temp_ptr, cmd);
		cmd_free(cmd);
		return 1;
	}
	if (lc_strncmp(cmd->command, "pass", -1) == 0) {
		temp_pass(temp_ptr, cmd);
		cmd_free(cmd);
		return 1;
	}
	if (lc_strncmp(cmd->command, "capab", -1) == 0) {
		temp_capab(temp_ptr, cmd);
		cmd_free(cmd);
		return 1;
	}

	/* Not registered yet error message */
	ircd_sprintf(buffer, BUF_SIZE, "451 %s :You have not registered",
		     (temp_ptr->nick[0] == '\0') ? "*" : temp_ptr->nick);
	send_temp(temp_ptr, buffer);

	cmd_free(cmd);
	return 0;
}

/* 
* register a local client 
*/
static int ircd_client_register(struct tempinfo *temp)
{
	ClientInfo client = NULL;
	char buffer[BUF_SIZE];
	char mode[10];
	char *serv_pass = NULL;
	int i;

	/* Check we have the flags we need */
	if (!(temp->flags & USR_NICK) || !(temp->flags & USR_USER)
	    || !(temp->flags & USR_RESOLVED)) {
		return 0;
	}

	/* Now, we can actually move them */
	/* Remove from the fd_list */
	serv_pass = fd_list[temp->fd].temp->pass;
	fd_list[temp->fd].temp = NULL;

	/* Allocate full clientinfo */
	client = ircd_malloc(sizeof(struct _clientinfo));

	/* Allocate memory for client list */
	client->chanlist_no = 10;
	client->chan_list = calloc(10, sizeof(ChannelInfo));
	if (client->chan_list == NULL) {
		log_msg(LOG_DEBUG1,
			"Failed to allocate space for channel list");
		close(temp->fd);
		return 0;
	}
	/* Zero out chan list */
	for (i = 0; i < client->chanlist_no; i++) {
		client->chan_list[i] = NULL;
	}

	/* Setup newly allocated memory */
	client->fd = temp->fd;
	client->mode = temp->mode;
	client->flags = temp->flags;
	client->ts = time(NULL);
	strncpy(client->nick, temp->nick, NICK_LENGTH);
	client->nick[NICK_LENGTH] = '\0';
	strncpy(client->host, temp->host, HOST_LENGTH);
	client->host[HOST_LENGTH] = '\0';
	strncpy(client->user, temp->user, USER_LENGTH);
	client->user[USER_LENGTH] = '\0';
	strncpy(client->real_name, temp->real_name, REAL_NAME_LENGTH);
	client->real_name[REAL_NAME_LENGTH] = '\0';
	client->lastrecv = temp->lastrecv;
	client->idle_timer = time(NULL);
	client->server = get_my_server();

	/* Add into tree */
	client_add_by_struct(client);
	get_my_server()->client_count++;
	/* Set pointer in the fd_list */
	fd_list[client->fd].client = client;

	/* Inherit SSL stuff */
#ifdef USE_OPENSSL
	if(temp->ssl) {
		/* XXX - This isn't right */
		client->ssl = temp->ssl;
		/* Set trusted mode for client */
		client->mode |= MODE_TRUSTED;
	}
#endif	

	/* Remove temp from the list */
		temp = NULL;
	temp_rem(client->fd);

	/* Send info to client */
	ircd_sprintf(buffer, BUF_SIZE,
		     ":Welcome to the Internet Relay Network %s",
		     client->nick);
	send_numeric(client, client->server, 1, buffer);
	ircd_sprintf(buffer, BUF_SIZE,
		     ":Your host is %s, running version %s%s",
		     client->server->name, VERSION_TAG, VERSION);
	send_numeric(client, client->server, 2, buffer);
	ircd_sprintf(buffer, BUF_SIZE, ":This server was created on %s",
		     server_config.load_date);
	send_numeric(client, client->server, 3, buffer);
	ircd_sprintf(buffer, BUF_SIZE, "%s %s%s %s %s",
		     client->server->name, VERSION_TAG, VERSION, USERMODES,
		     CHANMODES);
	send_numeric(client, client->server, 4, buffer);

	/* Send LUSERS replies */
	send_lusers(client, "*");

	/* Send MOTD */
	send_motd(client);

	/* Set and report mode */
	if (client->mode != 0) {
		/* Find out which modes are set */
		switch (client->mode & (MODE_INVIS | MODE_WALLOPS)) {
		case (MODE_INVIS | MODE_WALLOPS):
			ircd_sprintf(mode, 10, "iw");
			ircd_sprintf(buffer, 200, "MODE %s :%s",
				     client->nick, mode);
			send_client_client(client, client, buffer);
			break;

		case (MODE_INVIS):
			ircd_sprintf(mode, 10, "i");
			ircd_sprintf(buffer, 200, "MODE %s :+%s",
				     client->nick, mode);
			send_client_client(client, client, buffer);
			break;

		case (MODE_WALLOPS):
			ircd_sprintf(mode, 10, "w");
			ircd_sprintf(buffer, 200, ":%s  MODE %s :+%s",
				     client->nick, client->nick, mode);
			send_client_client(client, client, buffer);
			break;

		default:
			break;
		}
	}

	/* Inform servers */
	ircd_sprintf(buffer, BUF_SIZE, "NICK %s 0 %ld %s %s %s %s 0 :%s",
		     client->nick, client->ts, get_mode_str(client->mode),
		     client->user, client->host, client->server->name,
		     client->real_name);
	servers_announce(NULL, buffer);
	
	/* Forward the services */
	if(serv_pass) {
		ServerInfo services=NULL;

		services = server_locate(server_config.services_name);

		ircd_sprintf(buffer,BUF_SIZE,"PRIVMSG %s@%s :SIDENTIFY %s",
			server_config.services_nickserv,
			server_config.services_name,
			serv_pass);
		
		send_origin_server(services,client->nick,buffer);

		free(serv_pass);
	}

	return 1;
}

/* 
 * Register a server
 */
static int ircd_server_register(struct tempinfo *t_ptr)
{
	char buffer[BUF_SIZE];
	ServerInfo server = NULL;

	/* Check Flags */
	/* Only allow clients which haven't used USER/NICK Commands */
	log_msg(LOG_DEBUG2, "Attempting to register a server on fd %d",
		t_ptr->fd);
	if (!(t_ptr->flags & USR_SERVER) || !(t_ptr->flags & USR_PASS)
	    || (t_ptr->flags & USR_NICK) || (t_ptr->flags & USR_USER)) {
		return 0;
	}

	fd_list[t_ptr->fd].temp = NULL;

	/* Allocate memory for new server */
	server = ircd_malloc(sizeof(struct _serverinfo));

	/* Copy stuff into new structure */
	server->hops = 1;
	server->fd = t_ptr->fd;
#ifdef USE_OPENSSL
	if(t_ptr->enc_ctx) {
		server->enc_ctx = malloc(sizeof(struct _encryption_contexts));
		memcpy(server->enc_ctx, t_ptr->enc_ctx,sizeof(struct _encryption_contexts));
	}
#endif
	strncpy(server->name, t_ptr->name, SERVNAME_MAX);
	server->lastrecv = t_ptr->lastrecv;
	server->via_ptr = get_my_server();
	/* strncpy seems to cock some stuff up when used here */
	ircd_sprintf(server->info, 201, "%s", t_ptr->info);

	/* Add to server hash */
	if (!server_hash_add(server)) {
		log_msg(LOG_BASIC, "Failed to add server to hash");
		free(fd_list[t_ptr->fd].server);
		return 0;
	}

	/* Update the counters */
	get_my_server()->server_count++;

	/* Send them our information */
	if (!(t_ptr->flags & USR_SENTMYINFO)) {
		ircd_sprintf(buffer, BUF_SIZE, "SERVER %s 0 0 :%s",
			     get_my_server()->name, get_my_server()->info);
		send_origin_server(server, NULL, buffer);
	}

	/* Update fd list */
	fd_list[server->fd].server = server;

	/* Free temporary Structure */
	temp_rem(t_ptr->fd);
	t_ptr = NULL;

	/* Send SVINFO */
	ircd_sprintf(buffer, BUF_SIZE, "SVINFO %d %d %d :%ld", 4, 4, 1,
		     time(NULL));
	send_origin_server(server, NULL, buffer);

	/* Log it */
	log_msg(LOG_BASIC, "New Local Server Connection to %s - %s",
		server->name, server->info);

	/* Introduce the servers we know */
	servers_introduce(server);

	/* Introduce the clients we know */
	clients_introduce(server);

	/* introduce the channels using NJOIN */
	channels_introduce(server);

	/* Relay registration to other servers */
	ircd_sprintf(buffer, BUF_SIZE, "SERVER %s 0 0 :%s", server->name,
		     server->info);
	servers_relay(server, get_my_server()->name, buffer);

	ircd_sprintf(buffer, BUF_SIZE, "NetBurst completed with server %s",
		     server->name);
	clients_servnotice(buffer);

	return 1;
}


/* 
* Zeros everything out, and calls various init stuff 
*/
static int ircd_init(struct bind_info server_sockets[], fd_set * connections,
		     fd_set * serv_connects)
{
	unsigned int i;
	struct rlimit rlim;

	/* Initialise Resolver */
	resolver_init();

	/* Initialise server hash table */
	server_hash_init(server_config.server_name, server_config.server_info);

	assert(server_locate(server_config.server_name) != NULL);

	/* allocate the table */
	assert(getrlimit(RLIMIT_NOFILE, &rlim) == 0);

	max_fds = rlim.rlim_max;

	fd_list = calloc(max_fds, sizeof(struct fd_wrap));
	assert(fd_list != NULL);

	/* Initialise our array of fd's */
	for (i = 0; i < max_fds; i++) {
		fd_list[i].server = NULL;
		fd_list[i].client = NULL;
		fd_list[i].temp = NULL;
		fd_list[i].res_id = -1;
	}

	/* Set highest fd */
	for (i = 0; i < 32 && server_sockets[i].fd != -1; i++) {
		highest_fd = server_sockets[i].fd;
	}
	/* Zero out fd_set */
	FD_ZERO(connections);
	FD_ZERO(serv_connects);
	for (i = 0; i < 32 && server_sockets[i].fd != -1; i++) {
		FD_SET(server_sockets[i].fd, connections);
	}

	return 1;
}

/* 
* Check Pings 
*/
static int ircd_check_ping(struct fd_wrap *list, int i)
{
	time_t curr_lastrecv = 0;
	int *flags = NULL;


	if (list[i].temp != NULL || list[i].client != NULL
	    || list[i].server != NULL) {
		if (list[i].temp) {
			curr_lastrecv = list[i].temp->lastrecv;
			flags = &(list[i].temp->flags);
		}
		if (list[i].client) {
			curr_lastrecv = list[i].client->lastrecv;
			flags = &(list[i].client->flags);
		}
		if (list[i].server) {
			curr_lastrecv = list[i].server->lastrecv;
			flags = &(list[i].server->flags);
		}

		/* We have a client/server on this fd */
		if (curr_lastrecv + (2 * PING_SECS) < time(NULL) &&
		    (*flags & USR_PINGSENT)) {
			/* Kill the client, more than 180 secs idle */
			force_quit(i, "Ping Timeout");
		} else {
			/* Check to see if we need to ping them */
			if (curr_lastrecv + PING_SECS < time(NULL)
			    && !(*flags & USR_PINGSENT)) {
				/* Send a ping */
				*flags |= USR_PINGSENT;
				
				send_ping(ircd_fdlist_get_entry(i));
			}
		}
	}

	return 1;

}

/* 
* Check resolver status on every socket 
* Copy if complete 
*/
static int ircd_check_resolver(struct fd_wrap *wrapper)
{
	char *hostname = NULL;

	if (wrapper->res_id != -1) {
		hostname = resolver_check(wrapper->res_id);
		if (hostname) {
			/* We only need to use temp to resolve */
			/* Anything else should be resolved */
			if (wrapper->temp) {
				if (wrapper->temp->flags & USR_SHOULDCONN) {
					server_init_connect(wrapper->temp, hostname);				
				} else {
					/* Copy Hostname in */
					strncpy(wrapper->temp->host,hostname, HOST_LENGTH);
					/* Tell Client */
					if(wrapper->temp->flags & USR_SENTINITAUTH) {
						send_temp(wrapper->temp,"NOTICE * :*** Got Your Hostname"); 
					} else {
						send_temp(wrapper->temp,"NOTICE * :*** Looking Up Hostname"); 
						send_temp(wrapper->temp,"NOTICE * :*** Got Your Hostname"); 
						wrapper->temp->flags |= USR_SENTINITAUTH;
					}
					
				}
				wrapper->temp->flags |= USR_RESOLVED;
			}
			/* Not sure we need it */
			if (wrapper->server) {
				strncpy(wrapper->server->host, hostname,
					HOST_LENGTH);
			}
			wrapper->res_id = -1;
		}

	}

	return 1;
}

/*
* Relay numerics to the destination client
* Only relay 100+ in codes
*/
static int ircd_numeric_relay(Cmd cmd)
{
	int code, i;
	ClientInfo d_ptr = NULL;
	ServerInfo s_ptr = NULL;
	/* Use the real max size now */
	char buffer[512];

	/* Just return if its got no origin */
	if (!cmd->origin) {
		cmd_free(cmd);
		return 0;
	}

	/* Check it is from a server */
	s_ptr = server_locate(cmd->origin);
	if (!s_ptr) {
		cmd_free(cmd);
		return 0;
	}

	/* Check the numeric is between 100 and 999 inclusive */
	code = atoi(cmd->command);
	if (code < 100 || code > 999) {
		cmd_free(cmd);
		return 0;
	}

	/* Now get the dest */
	d_ptr = client_locate(cmd->params[0]);

	if (!d_ptr) {
		cmd_free(cmd);
		return 0;
	}

	ircd_sprintf(buffer, 490, "%i %s", code, cmd->params[0]);
	for (i = 1; cmd->params[i]; i++) {
		ircd_sprintf(buffer, 490, "%s %s", buffer, cmd->params[i]);
	}
	ircd_sprintf(buffer, 490, "%s :%s", buffer, cmd->text);

	send_origin_client(d_ptr, s_ptr->name, buffer);

	cmd_free(cmd);
	return 1;
}

/*
 * Remove this entry from the fd table and clean out the
 * highest fd if we need to
 */
void ircd_fdlist_clear_entry(int is_highest)
{
	int i;

	if (is_highest < 0)
		return;

	fd_list[is_highest].client = NULL;
	fd_list[is_highest].server = NULL;
	fd_list[is_highest].temp = NULL;
	fd_list[is_highest].res_id = -1;

	if (is_highest == highest_fd) {
		/* Only do this if we need to */
		for (i = highest_fd; i >= 0; i--) {
			if (fd_list[i].client || fd_list[i].server
			    || fd_list[i].temp) {
				highest_fd = i;
				i = -3;
			}
		}
		if (i == 0) {
			/* Can't use the table, let the main loop
			 * revert to the server sockets */
			highest_fd = -1;
		}
	}


}

/*
 * Get an fd_wrap * for a particular fd
 */
struct fd_wrap *ircd_fdlist_get_entry(int fd)
{
	if (fd < 0)
		return NULL;

	return &fd_list[fd];
}

/*
 * Add an entry to the fdlist
 */
int ircd_fdlist_set_entry(int fd, struct fd_wrap *new_contents)
{
	if (fd >= 0) {
		memcpy(&(fd_list[fd]), new_contents,
		       sizeof(struct fd_wrap));

		if (fd > highest_fd) {
			highest_fd = fd;
		}

		return 0;
	} else {
		return -1;
	}
}

