/* This file is part of Chunky Monkey IRCD 
 *
 * Chunky Monkey IRCD is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Chunky Monkey IRCD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Chunky Monkey IRCD; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * $Id: modules.c,v 1.28.4.4 2002/05/18 11:24:23 lloydy Exp $
 *
 */
#include "modules.h"


/* Modules.c - Handles the loading/unloading of command
 * modules uses dlopen/dlsym/dlclose.  Automatically
 * loads modules in $prefix/lib/ircd/modules, and then
 * disables modules as directed in the config file
 *
 */

#ifdef USING_OPENBSD
int module_match(struct dirent *entry)
{
	return check_string(entry->d_name, "*.so");
}
#else
int module_match(const struct dirent *entry)
{
	return check_string(entry->d_name, "*.so");
}
#endif


static int module_load(ModInfo, ClientInfo, ServerInfo, Cmd);
static int module_init_check(ModInfo);

/* List Start */
static ModInfo list_start = NULL;
/* Signal Handling */
extern int signal_caught;
/* Segmentation Fault Protection */
ModInfo in_module = NULL;
jmp_buf env;

/*
 * Check to see if we can init a module
 * Also gets version number and flags
 */
static int module_init_check(ModInfo init)
{
	char path[200];
	void *handle = NULL;
	int (*sym) (ClientInfo, ServerInfo, Cmd) = NULL;
	char full_name[30];
	float *version;
	int *flags;

#ifdef BROKEN_DLOPEN
	ircd_sprintf(full_name, 30, "_cmd_%s", init->name);
#else
	ircd_sprintf(full_name, 30, "cmd_%s", init->name);
#endif
	ircd_sprintf(path, 200, "%s/cmd_%s.so", MODDIR, init->name);

	handle = dlopen(path, DL_FLAGS);

	if (!handle) {
		return 0;
	} else {
		sym =
		    (int (*)(ClientInfo, ServerInfo, Cmd)) dlsym(handle,
								 full_name);
#ifdef BROKEN_DLOPEN
		flags = (int *) dlsym(handle, "_flags");
		version = (float *) dlsym(handle, "_version");
#else
		flags = (int *) dlsym(handle, "flags");
		version = (float *) dlsym(handle, "version");
#endif
		if (!sym || !flags || !version) {
			dlclose(handle);
			return 0;
		}
	}
	init->flags = *flags;
	init->version = *version;
	dlclose(handle);

	return 1;
}



/*
 * Initalise the module list
 */
int ircd_module_init()
{
	int entries_count = 0;
	struct dirent **namelist;
	ModInfo curr = NULL;
	ModInfo prev = NULL;
	ModInfo init = NULL;
	char *name = NULL;
	int i, fine;

	entries_count = scandir(MODDIR, &namelist, module_match, 0);

	if (entries_count < 0) {
		log_msg(LOG_BASIC, "No modules in module directory (%s)",
			MODDIR);
		signal_caught = 1;
	} else {
		while (entries_count--) {
			name = NULL;
			fine = 0;
			init = ircd_malloc(sizeof(struct _modinfo));

			name = namelist[entries_count]->d_name;

			for (i = strlen(name); i > 0; i--) {
				if (name[i] == '.')
					name[i] = '\0';
				if (name[i] == '_') {
					fine = 1;
					name = &(name[i + 1]);
				}

			}

			strncpy(init->name, name, 30);
			if (fine == 0 || module_init_check(init) == 0) {
				log_msg(LOG_BASIC,
					"Disabled module %s (cmd_%s.so) - %s",
					name, name,
					fine ? "Symbol Unavailable" :
					"Wrong Filename");
				free(init);
				continue;
			}

			curr = list_start;
			prev = list_start;


			log_msg(LOG_BASIC,
				"Enabling module %s (%s/cmd_%s.so)",
				init->name, MODDIR, init->name);
			log_msg(LOG_BASIC, "Module %s has version %.2f",
				init->name, init->version);
			while (curr) {
				prev = curr;
				curr = curr->next;
			}
			if (prev) {
				prev->next = init;
			} else {
				list_start = init;
			}

			free(namelist[entries_count]);
		}
		free(namelist);
	}

	return 1;
}

/*
 * Called by REHASH to reload the modules
 */
int ircd_module_rehash()
{
	ModInfo curr = NULL, prev = NULL;

	curr = list_start;

	log_msg(LOG_BASIC, "Rehashing Module List");

	while (curr) {
		prev = curr;
		curr = curr->next;
		free(prev);
		prev = NULL;
	}

	list_start = NULL;

	/* Reload The Modules */
	ircd_module_init();

	return 1;
}

/*
 * Unload a module or all the modules when requested
 * by an oper
 */
int cmd_unload(ClientInfo c_ptr, ServerInfo s_ptr, Cmd cmd_ptr)
{
	char buffer[BUF_SIZE];

	/* Only accept this command from local clients */
	if (!c_ptr || s_ptr) {
		log_msg(LOG_DEBUG2,
			"Got an UNLOAD command from a non-local client");
		return 1;
	}

	/* Check for operator status */
	if (!(c_ptr->mode & MODE_IRCOP)) {
		log_msg(LOG_DEBUG2, "Got an UNLOAD command from a non-op");
		send_numeric(c_ptr, get_my_server(), 481,
			     ":You're no operator");
		return 1;
	}

	/* Now check the sanity of the parameter */
	if (cmd_ptr->params[0] != NULL) {
		if (lc_strncmp(cmd_ptr->params[0], "*", -1) == 0) {
			ircd_module_unload_all();
			send_numeric(c_ptr, get_my_server(), 271,
				     ":All modules unloaded");
		} else {
			if (ircd_module_unload(cmd_ptr->params[0])) {
				ircd_sprintf(buffer, BUF_SIZE,
					     "%s :Module unloaded",
					     cmd_ptr->params[0]);
				send_numeric(c_ptr, get_my_server(), 270,
					     buffer);
			} else {
				ircd_sprintf(buffer, BUF_SIZE,
					     "%s :No Such Module",
					     cmd_ptr->params[0]);
				send_numeric(c_ptr, get_my_server(), 503,
					     buffer);
			}

		}
	} else {
		log_msg(LOG_DEBUG1,
			"Tried to unload a non-existant module");

	}

	return 1;
}

/* 
 * Unload a module
 */
int ircd_module_unload(char *name)
{
	ModInfo curr = NULL;

	curr = list_start;

	while (curr) {
		if (!lc_strncmp(curr->name, name, -1)) {
			curr->use_count = 0;
			if (curr->handle)
				dlclose(curr->handle);
			curr->handle = NULL;
			return 1;
		} else {
			curr = curr->next;
		}
	}

	return 0;
}

/* 
 * Unload all the modules
 */
int ircd_module_unload_all()
{
	ModInfo curr = NULL;

	curr = list_start;

	while (curr) {
		curr->use_count = 0;
		if (curr->handle) {
			dlclose(curr->handle);
			curr->handle = NULL;
		}
		curr = curr->next;
	}

	return 1;
}

/*
 * Disable a module
 */
int ircd_module_disable(char *name)
{
	ModInfo curr = NULL, prev = NULL;

	curr = list_start;

	while (curr) {
		if (!lc_strncmp(curr->name, name, -1)) {
			if (prev) {
				prev->next = curr->next;
				free(curr);
			} else {
				list_start = curr->next;
				free(curr);
			}
			log_msg(LOG_BASIC,
				"Disabled module %s (cmd_%s.so)", name,
				name);
			return 1;
		} else {
			prev = curr;
			curr = curr->next;
		}
	}

	log_msg(LOG_BASIC, "Unable to disable module %s (cmd_%s.so)", name,
		name);
	return 0;
}

/*
 * Called from ircd.c to execute the correct routine
 * Returns 1 if we find the right module, 0 otherwise
 */
int ircd_module_dlopen(ClientInfo c_ptr, ServerInfo s_ptr, Cmd cmd)
{
	ModInfo curr = NULL;
	int done = 0;

	curr = list_start;

	while (!done) {
		if (curr) {
			if (!lc_strncmp(curr->name, cmd->command, -1)) {
				/* This is the command we want */
				return module_load(curr, c_ptr, s_ptr,
						   cmd);
			} else {
				curr = curr->next;
			}
		} else {
			done = 1;
		}

	}

	/* Failed to match a command */
	return 0;
}

/*
 * Actually load the module, check the dlopen() stuff
 */
static int module_load(ModInfo module,
		       ClientInfo c_ptr, ServerInfo s_ptr, Cmd cmd)
{
	char name[30] = { '\0' };
	char path[200] = { '\0' };
	int (*ptr) (ClientInfo, ServerInfo, Cmd) = NULL;

	/* Check the permissions for the module */
	if ((c_ptr && !(module->flags & CMD_CLIENT)) ||	/* Client Command */
	    (s_ptr && !(module->flags & CMD_SERVER))	/* Server COmmand */
	    ) {
		log_msg(LOG_DEBUG2, "Refused request to load module %s",
			module->name);
		return 0;
	}
#ifdef BROKEN_DLOPEN
	ircd_sprintf(name, 30, "_cmd_%s", module->name);
#else
	ircd_sprintf(name, 30, "cmd_%s", module->name);
#endif

	if (!module->handle) {
		/* Get a handle */
		ircd_sprintf(path, 200, "%s/cmd_%s.so", MODDIR,
			     module->name);

		module->handle = dlopen(path, DL_FLAGS);

		if (!module->handle) {
			log_msg(LOG_BASIC, "Unable to open module %s",
				path);
			signal_caught = 0;
			return 0;
		}

	}
	/* Get the symbol */
	ptr =
	    (int (*)(ClientInfo, ServerInfo, Cmd)) dlsym(module->handle,
							 name);

	if (!ptr) {
		log_msg(LOG_BASIC, "Unable to get symbol %s", name);
		signal_caught = 0;
		return 0;
	}

	/* Call the routine */
	module->use_count++;

	in_module = module;
	if (setjmp(env) == 0) {
		ptr(c_ptr, s_ptr, cmd);
	}
	in_module = NULL;


	return 1;


}

/* 
 * Segfault handler
 * If the segfault is in a module, attempt to restore context 
 * just before the call, and disable the module that caused the segfault
 * In -d 1 and above, we abort() immediately
 */
void caught_segfault(int signal)
{
	char name[30];

	if (signal == SIGSEGV) {
		if (in_module) {
			if (!server_config.debug) {
				log_msg(LOG_BASIC,
					"Caught a segmentation fault in module %s, attempting to continue",
					in_module->name);
				strncpy(name, in_module->name, 30);
				in_module = NULL;
				ircd_module_disable(name);
				longjmp(env, -1);
			} else {
				log_msg(LOG_BASIC,
					"Caught a segmentation fault in module %s, calling abort()",
					in_module->name);
				abort();
			}
		} else {
			/* Bug in the IRCD */
			log_msg(LOG_BASIC,
				"Segmentation fault in the IRCD, aborting");
			abort();
		}
	}
}

/*
* Rehash - Reload config file
*/
int cmd_rehash(ClientInfo c_ptr, ServerInfo s_ptr, Cmd cmd)
{

	char buffer[200];

	cmd = cmd;		/* Avoid not used error */

	if (s_ptr) {

		return 0;
	}

	if (!(c_ptr->mode & MODE_IRCOP)) {
		send_numeric(c_ptr, get_my_server(), 481,
			     ":Permission Denied- You're no IRC Op");

		return 0;
	}

	ircd_sprintf(buffer, 200, "%s :Rehashing", server_config.conf_file);
	send_numeric(c_ptr, get_my_server(), 382, buffer);
	rehash(SIGHUP);


	return 1;
}
