/* This file is part of Chunky Monkey IRCD 
 *
 * Chunky Monkey IRCD is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Chunky Monkey IRCD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Chunky Monkey IRCD; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * $Id: cmd_sjoin.c,v 1.23.4.5 2002/06/03 20:03:55 lloydy Exp $
 *
 */
#include "modules.h"

CMIRCD_MODULE(1.02, CMD_SERVER);

/* Internal Routines */
static int sjoin_ts_older(ServerInfo, ChannelInfo, Cmd);
static int sjoin_ts_equal(ServerInfo, ChannelInfo, Cmd);
static int sjoin_ts_newer(ServerInfo, ChannelInfo, Cmd);
static int sjoin_join(ChannelInfo, char *, int);
static char *sjoin_mode(ChannelInfo, Cmd);
static char *sjoin_get_old_mode(ChannelMember);
static int sjoin_remove_all_modes(ChannelInfo);
static int sjoin_scan_local_ops(ChannelInfo);
static int sjoin_scan_sjoin_ops(Cmd);
static int sjoin_relay(ServerInfo, time_t, char *, char *, char *, int);

/*
* SJOIN - Used to transfer channel information over server links
* cmd_sjoin, check timestamp, and call correct routine to deal with it
*/
int cmd_sjoin(ClientInfo c_ptr, ServerInfo s_ptr, Cmd cmd)
{
	time_t remote_ts = 0;
	ChannelInfo chan = NULL;

	/* Only Allow SJOIN over server links */
	if (c_ptr) {

		return 0;
	}

	/* Check we have enough params */
	if (!check_params(cmd, 3)) {

		return 0;
	}

	/* Find out what to do */
	remote_ts = strtoul(cmd->params[0], NULL, 10);
	chan = locate_channel(cmd->params[1]);
	if (chan || (chan = channel_create(cmd->params[1], remote_ts))) {
		if (remote_ts < chan->ts) {
			sjoin_ts_older(s_ptr, chan, cmd);
		} else {
			if (remote_ts > chan->ts) {
				sjoin_ts_newer(s_ptr, chan, cmd);
			} else {
				if (remote_ts == chan->ts) {
					sjoin_ts_equal(s_ptr, chan, cmd);
				}
			}
		}
	}



	return 1;

}

/*
* sjoin_ts_older()
* the mode from the other server is older than ours
*/
static int sjoin_ts_older(ServerInfo s_ptr, ChannelInfo chan, Cmd cmd)
{
	int local_ops = 0;
	int remote_ops = 0;

#ifdef DEBUG
	printf("Using sjoin_ts_older\n");
	printf("On channel %s\n", chan->name);
#endif

	local_ops = sjoin_scan_local_ops(chan);
	remote_ops = sjoin_scan_sjoin_ops(cmd);

	/* If SJOIN ops someone */
	if (remote_ops) {
		/* Remove all modes */
		sjoin_remove_all_modes(chan);
		/* Set our TS to remote one */
		chan->ts = strtoul(cmd->params[0], NULL, 10);
		/* Add users locally */
		sjoin_join(chan, cmd->text, 1);
		/* Relay */
		sjoin_relay(s_ptr, chan->ts, sjoin_mode(chan, cmd),
			    cmd->text, chan->name, 1);
		/* Return */
		return 1;
	}

	/* If the SJOIN doesn't op anyone, and we have ops */
	/* Spread with our own ts */
	if (!remote_ops && local_ops) {
		sjoin_join(chan, cmd->text, 0);
		sjoin_relay(s_ptr, chan->ts, sjoin_mode(chan, cmd),
			    cmd->text, chan->name, 1);
		return 1;
	}

	/* Otherwise, just act on it, and send */
	sjoin_join(chan, cmd->text, 0);
	sjoin_relay(s_ptr, strtoul(cmd->params[0], NULL, 10),
		    sjoin_mode(chan, cmd), cmd->text, chan->name, 1);


	return 0;



}

/*
* sjoin_ts_newer()
* The modes from the other server are newer than ours
* (we ignore them)
*/
static int sjoin_ts_newer(ServerInfo s_ptr, ChannelInfo chan, Cmd cmd)
{
	int remote_ops = 0, local_ops = 0;

#ifdef DEBUG
	printf("Using sjoin_ts_newer\n");
	printf("On channel %s\n", chan->name);
#endif

	/* Look at the situation */
	remote_ops = sjoin_scan_sjoin_ops(cmd);
	local_ops = sjoin_scan_local_ops(chan);

	/* If the SJOIN contains no ops */
	if (!remote_ops) {
		/* Add Users */
		sjoin_join(chan, cmd->text, 0);
		/* Relay with our TS */
		sjoin_relay(s_ptr, chan->ts, NULL, cmd->text, chan->name,
			    0);
		/* Return */
		return 1;
	}

	/* If we have ops */
	/* Relay without ops */
	if (remote_ops && local_ops) {
		/* Add Users */
		sjoin_join(chan, cmd->text, 0);
		/* Relay with our ts */
		sjoin_relay(s_ptr, chan->ts, NULL, cmd->text, chan->name,
			    0);
		/* Return */
		return 1;
	}

	/* If there are ops, and we don't have any */
	if (remote_ops && !local_ops) {
		/* Add with ops */
		sjoin_join(chan, cmd->text, 1);
		/* Use remote ts */
		chan->ts = strtoul(cmd->params[0], NULL, 10);
		/* relay with ops and the ts we recieved */
		sjoin_relay(s_ptr, chan->ts, NULL, cmd->text, chan->name,
			    1);
		/* Return */
		return 1;
	}


	return 0;
}

/*
* sjoin_ts_equal()
* The data we receive is from the same time as
* our data, probably correct from before the netsplit
*/
static int sjoin_ts_equal(ServerInfo s_ptr, ChannelInfo chan, Cmd cmd)
{
#ifdef DEBUG
	printf("Using sjoin_ts_equal\n");
#endif

	/* Grab the channel mode */
	sjoin_mode(chan, cmd);

	/* Add Users in */
	sjoin_join(chan, cmd->text, 1);

	/* Relay */
	sjoin_relay(s_ptr, chan->ts, NULL, cmd->text, chan->name, 1);

	return 1;
}

/*
* sjoin_join(char *text,int allow_modes)
* Takes cients from text, adds them to the channel
* sends appropriate message to the local clients, and changes mode if required
*/
static int sjoin_join(ChannelInfo chan, char *r_users, int allow_modes)
{
	int i = 0, j = 0, last = 0;
	int mode = 0, done = 0;
	ClientInfo client = NULL;
	int str_len = 0;
	char *temp = NULL, *nick = NULL;
	char buffer[BUF_SIZE];
	char *users = NULL;

	/* 
	 * Return if we don't have a r_users
	 */
	if(!r_users) 
		return 0;
	

	str_len = strlen(r_users);
	users = string_alloc(str_len);
	strcpy(users, r_users);

	for (i = 0; i <= str_len; i++) {
		/* Get a client */
		if (users[i] == ' ' || users[i] == '\0') {
			users[i] = '\0';
			temp = &(users[last]);
			i++;
			last = i;
			mode = 0;
			done = 0;
			client = 0;
			nick = 0;
			j = 0;

			/* temp now points to a client nick */
			/* Get the mode */
			for (j = 0, done = 0; done == 0; j++) {
				switch (temp[j]) {
				case '+':
					mode |= CHANUSR_VOICE;
					break;

				case '@':
					/* XXX - Creator Support? */
					mode |= CHANUSR_OP;
					break;

				case '%':
					mode |= CHANUSR_HALFOP;
					break;

				default:
					done = 1;
					break;
				}
			}
			nick = &(temp[--j]);
			client = client_locate(nick);

			/* Got a client, now add to the channel lists */
			if ((client)
			    && !(channel_is_client_on(client, chan))) {
				client_add_chan(client, chan);

				/* Check Mode Stuff */
				/* Add to channel list with mode */
				if (allow_modes) {
					channel_add_client(client, chan,
							   mode);
				} else {
					channel_add_client(client, chan,
							   0);
				}

				/* Inform channel of JOIN */
				ircd_sprintf(buffer, BUF_SIZE, "JOIN %s",
					     chan->name);
				send_to_local_channel(client, chan,
						      buffer);

				/* Inform channel of MODE */
				if (allow_modes && mode) {
					char str_mode[5] = "\0\0\0\0\0";
					char params[200] = "\0";
					int pos = 1;

					str_mode[0] = '+';

					if (mode & CHANUSR_OP) {
						str_mode[pos++] = 'o';
						strncat(params,
							client->nick,
							200 -
							strlen(params));
						strncat(params, " ",
							200 -
							strlen(params));
					}
					if (mode & CHANUSR_HALFOP) {
						str_mode[pos++] = 'h';
						strncat(params,
							client->nick,
							200 -
							strlen(params));
						strncat(params, " ",
							200 -
							strlen(params));
					}
					if (mode & CHANUSR_VOICE) {
						str_mode[pos++] = 'v';
						strncat(params,
							client->nick,
							200 -
							strlen(params));
						strncat(params, " ",
							200 -
							strlen(params));
					}
					str_mode[pos] = '\0';

					ircd_sprintf(buffer, BUF_SIZE,
						     ":%s MODE %s %s %s",
						     client->server->name,
						     chan->name, str_mode,
						     params);

					send_to_local_channel(NULL, chan,
							      buffer);

				}
			}

		}
	}
	free(users);
	return 1;
}

/*
* sjoin_mode(ChannelInfo chan,Cmd cmd)
* used to extract the channel mode from a command, and set it
* only called if we should be, so no protection needed 
*/
static char *sjoin_mode(ChannelInfo chan, Cmd cmd)
{

	int limit = -1;
	char *key = NULL;
	int new_mode = 0;
	int curr_param = 3;
	int str_len;
	int mode_pos = 1;
	char mode[20];
	char params[200];
	static char modeline[300];
	char buffer[BUF_SIZE];
	int i;

	mode[0] = '+';
	params[0] = '\0';

	str_len = strlen(cmd->params[2]);

	if (cmd->params[2][0] == '0') {
		strcpy(modeline, "0");
		return modeline;
	}

	for (i = 0; i < str_len; i++) {

		switch (cmd->params[2][i]) {

		case 'n':
			new_mode = CMODE_MEMONLYSEND;
			mode[mode_pos++] = 'n';
			break;

		case 't':
			new_mode = CMODE_TOPIC;
			mode[mode_pos++] = 't';
			break;

		case 's':
			new_mode = CMODE_SECRET;
			mode[mode_pos++] = 's';
			break;

		case 'm':
			new_mode = CMODE_MODERATED;
			mode[mode_pos++] = 'm';
			break;

		case 'l':
			limit = atoi(cmd->params[curr_param]);
			mode[mode_pos++] = 'l';
			ircd_sprintf(params, BUF_SIZE, "%ld", params,
				     limit);
			curr_param++;
			break;

		case 'k':
			key = cmd->params[curr_param];
			mode[mode_pos++] = 'k';
			ircd_sprintf(params, BUF_SIZE, "%s %s", params,
				     key);
			curr_param++;
			break;
		}
	}

	mode[mode_pos] = '\0';

	if (mode[1] != '\0') {

		/* Create Mode Line */
		ircd_sprintf(modeline, 300, "%s %s", mode, params);

		/* Relay Mode to local clients */
		ircd_sprintf(buffer, BUF_SIZE, ":%s MODE %s %s",
			     get_my_server()->name, chan->name, modeline);

		send_to_local_channel(NULL, chan, buffer);
	}


	return modeline;
}

/*
* Print a modeline to show what we've removed from a client
*/
static char *sjoin_get_old_mode(ChannelMember w_ptr)
{
	static char mode[200];

	if ((w_ptr->chan_mode & CHANUSR_OP) &&
	    (w_ptr->chan_mode & CHANUSR_VOICE)) {
		ircd_sprintf(mode, 200, "-ov %s %s",
			     w_ptr->user_info->nick,
			     w_ptr->user_info->nick);
		return mode;
	}

	if ((w_ptr->chan_mode & CHANUSR_OP) ||
	    (w_ptr->chan_mode & CHANUSR_VOICE)) {
		ircd_sprintf(mode, 200, "-%s %s",
			     w_ptr->chan_mode & CHANUSR_OP ? "o" : "v",
			     w_ptr->user_info->nick);
		return mode;
	}

	return NULL;
}

/*
* Remove all modes on the channel
* Called before we add remote clients
*/
static int sjoin_remove_all_modes(ChannelInfo chan)
{
	ChannelMember curr_ptr = NULL;
	char buffer[BUF_SIZE];

	/* Remove Channel Mode */
	ircd_sprintf(buffer, BUF_SIZE, ":%s MODE %s %s",
		     get_my_server()->name, chan->name,
		     get_chan_mode_str(chan, '-'));
	send_to_local_channel(NULL, chan, buffer);
	chan->mode = 0;


	curr_ptr = chan->users_start;

	/* Loop thru all clients */
	/* Remove their modes */
	while (curr_ptr) {
		if (curr_ptr->chan_mode != 0) {
			/* Inform local members of the change */
			ircd_sprintf(buffer, BUF_SIZE, ":%s MODE %s %s",
				     get_my_server()->name, chan->name,
				     sjoin_get_old_mode(curr_ptr));
			send_to_local_channel(NULL, chan, buffer);
			curr_ptr->chan_mode = 0;
		}
		curr_ptr = curr_ptr->next;
	}

	return 1;

}

/*
* Scan for local Ops
* Return 1 if we find one, 0 if not
*/
static int sjoin_scan_local_ops(ChannelInfo chan)
{
	ChannelMember curr_ptr = NULL;

	curr_ptr = chan->users_start;

	while (curr_ptr) {
		if (curr_ptr->chan_mode & CHANUSR_OP) {
			return 1;
		}

		curr_ptr = curr_ptr->next;
	}

	return 0;
}

/*
* Scan for mode changes in the SJOIN
* Return 1 if we find one, and 0 if not
*/
static int sjoin_scan_sjoin_ops(Cmd cmd)
{
	int i;

	for (i = 0; cmd->text[i] != '\0'; i++) {

		switch (cmd->text[i]) {
		case '@':
		case '+':
			return 1;
			break;

		default:
			break;
		}
	}

	return 0;
}

/*
* Create SJOIN for relaying to other servers
*/
static int sjoin_relay(ServerInfo s_ptr,
		       time_t ts, char *mode,
		       char *r_users, char *name, int use_modes)
{
	int str_len = 0, i = 0, j = 0;
	char *temp = NULL, *nick = NULL;
	int last = 0, done = 0;
	char buffer[BUF_SIZE];
	int new = 0;
	char *users = NULL;

	str_len = strlen(r_users);
	users = string_alloc(str_len);
	strcpy(users, r_users);

	ircd_sprintf(buffer, BUF_SIZE, "SJOIN %ld %s %s :", ts, name,
		     mode ? mode : "+ ");
	new = 1;

	for (i = 0; i <= str_len; i++) {
		/* Get a client */
		if (users[i] == ' ' || users[i] == '\0') {
			users[i] = '\0';
			temp = &(users[last]);
			i++;
			last = i;
			done = 0;
			nick = 0;
			j = 0;

			/* temp now points to a client nick */
			/* Get the mode */
			for (j = 0, done = 0; done == 0; j++) {
				switch (temp[j]) {
				case '@':
				case '+':
					break;

				default:
					done = 1;
					break;
				}
			}
			nick = &(temp[--j]);

			/* XXX - Might need to handle overflows, shouldn't happen */
			strcat(buffer, use_modes ? temp : nick);
			strcat(buffer, " ");
		}
	}

	servers_relay(s_ptr, get_my_server()->name, buffer);

	free(users);
	return 1;
}
